# -*- coding: utf-8 -*-
# mypy: disable-error-code="index"

"""
common type aliases, constants and helper functions
"""

from typing import Any, Iterable, TypeAlias


ScalarType: TypeAlias = str | int | float | bool | None
CollectionType: TypeAlias = dict | list
ValueType: TypeAlias = ScalarType | CollectionType

SegmentsTuple: TypeAlias = tuple[ScalarType, ...]
IndexType: TypeAlias = str | SegmentsTuple


COMMA_BLANK = ", "
DOT = "."
EMPTY = ""
DOUBLE_QUOTE = '"'
SLASH = "/"


def assured_collection(
    original: Any,
    error_message: str = "Expected a collection (ie. a dict or a list",
) -> CollectionType:
    """assure the original is a dict or a list"""
    if isinstance(original, (dict, list)):
        return original
    #
    raise TypeError(error_message)


def assured_number(
    original: Any, error_message: str = "Expected a number (float or int, not bool)"
) -> float | int:
    """assure the original is a int or float"""
    if not isinstance(original, bool) and isinstance(original, (float, int)):
        return original
    #
    raise TypeError(error_message)


def assured_scalar(
    original: Any,
    error_message: str = "Expected a scalar (ie. a string, an int, a float,"
    " or a bool), or None",
) -> ScalarType:
    """assure the original is a scalar"""
    if isinstance(original, (str, int, float, bool)) or original is None:
        return original
    #
    raise TypeError(error_message)


def assured_segments_tuple(
    original: Any,
    error_message: str = "Expected a segments tuple",
) -> SegmentsTuple:
    """assure the original is a segments tuple"""
    if isinstance(original, tuple):
        collector: list[ScalarType] = []
        for item in original:
            collector.append(assured_scalar(item, error_message=error_message))
        #
        return tuple(collector)
    #
    raise TypeError(error_message)


def assured_index(original: Any) -> IndexType:
    """assure the original is a valid index"""
    if isinstance(original, str):
        return original
    #
    return assured_segments_tuple(original, error_message="Expected a valid index")


def assured_dict(original: Any, error_message: str = "Expected a dict") -> dict:
    """assure the original is a dict"""
    if isinstance(original, dict):
        return original
    #
    raise TypeError(error_message)


def assured_float(original: Any, error_message: str = "Expected a float") -> float:
    """assure the original is a float"""
    if isinstance(original, float):
        return original
    #
    raise TypeError(error_message)


def assured_int(
    original: Any, error_message: str = "Expected an int (not bool)"
) -> int:
    """assure the original is an int"""
    if not isinstance(original, bool) and isinstance(original, int):
        return original
    #
    raise TypeError(error_message)


def assured_str(original: Any, error_message: str = "Expected a string") -> str:
    """assure the original is a str"""
    if isinstance(original, str):
        return original
    #
    raise TypeError(error_message)


def assured_list(original: Any, error_message: str = "Expected a list") -> list:
    """assure the original is a list"""
    if isinstance(original, list):
        return original
    #
    raise TypeError(error_message)


def assured_iterable(
    original: Any, error_message: str = "Expected an iterable"
) -> Iterable:
    """assure the original is an iterable"""
    if isinstance(original, Iterable):
        return original
    #
    raise TypeError(error_message)


def to_text(
    original: Any,
    encoding: str = "utf-8",
    error_message: str = "Expected a string or bytes",
) -> str:
    """Return a str from original.
    If the error message is left blank, convert any type to str
    """
    if isinstance(original, str):
        return original
    #
    if isinstance(original, (bytes, bytearray)):
        return original.decode(encoding)
    #
    if error_message:
        raise TypeError(error_message)
    #
    return str(original)


# vim: fileencoding=utf-8 ts=4 sts=4 sw=4 autoindent expandtab syntax=python:



# -*- coding: utf-8 -*-

"""
locks module
"""

from collections.abc import Callable
from threading import Condition, RLock
from time import monotonic as _time


_ZERO = 0
_EXCLUSIVE_LOCK = -1
_UNLOCKED = 0


class LockManager:
    """Context manager for BoltTypeLock classes"""

    def __init__(self, btl: "BoltTypeLock", shared: bool = True) -> None:
        """Register the acquire and release methods"""
        if shared:
            self.__acquire: Callable = btl.acquire_shared
            self.__release: Callable = btl.release_shared
        else:
            self.__acquire = btl.acquire_exclusive
            self.__release = btl.release_exclusive
        #

    def __enter__(self) -> bool:
        """acquire the lock"""
        return self.__acquire()

    def __exit__(self, t, v, tb):
        """release the lock"""
        self.__release()


class BoltTypeLock:
    """This class implements a bolt-type lock for exclusive or shared access,
    heavily borrowing from the threading.Semaphore class.

    BoltTypeLock instances maintain an internal value representing
    one of the following states:

        0  → the resource is available
        1  → the resource is exclusively locked
        2+ → the resource is in shared lock

    acquiring an exclusive lock: 0 → -1 only
    releasing an exclusive lock: -1 → 0
    acquiring a shared lock: increase the value if it is 0 or more
    releasing a shared lock: decrease the value if it is 1 or greater

    """

    def __init__(self) -> None:
        """Initialize the value"""
        self._cond = Condition(RLock())
        self.__state: int = _UNLOCKED
        self.shared = LockManager(self, shared=True)
        self.exclusive = LockManager(self, shared=False)

    @property
    def state(self) -> int:
        """observable state"""
        return self.__state

    def __repr__(self) -> str:
        """String rpresentation"""
        cls = self.__class__
        return (
            f"<{cls.__module__}.{cls.__qualname__} at {id(self):#x}:"
            f" state={self.__state}>"
        )

    def acquire_exclusive(self, timeout: float = _ZERO) -> bool:
        """Acquire an exclusive lock

        When invoked with a timeout other than 0, it will block for at
        most timeout seconds.  If acquire_exclusove does not complete successfully
        in that interval, return false.  Return true otherwise.

        """
        rc = False
        endtime: float = _ZERO
        with self._cond:
            while self.__state != _UNLOCKED:
                if timeout is _ZERO:
                    self._cond.wait(None)
                else:
                    if endtime is _ZERO:
                        endtime = _time() + timeout
                    elif _time() > endtime:
                        break
                    #
                    self._cond.wait(timeout)
                #
                # re-notify on continued wait
                self._cond.notify(1)
            else:
                self.__state = _EXCLUSIVE_LOCK
                rc = True
            #
        #
        return rc

    def acquire_shared(self, timeout: float = _ZERO):
        """Acquire a shared lock

        When invoked with a timeout other than 0, it will block for at
        most timeout seconds.  If acquire_shared does not complete successfully
        in that interval, return false.  Return true otherwise.

        """
        rc = False
        endtime: float = _ZERO
        with self._cond:
            while self.__state < 0:
                if timeout is _ZERO:
                    self._cond.wait(None)
                else:
                    if endtime is _ZERO:
                        endtime = _time() + timeout
                    elif _time() > endtime:
                        break
                    #
                    self._cond.wait(timeout)
                #
                # re-notify on continued wait
                self._cond.notify(1)
            else:
                self.__state += 1
                rc = True
            #
        #
        return rc

    def release_exclusive(self):
        """Release an exclusive lock"""
        with self._cond:
            if self.__state == _EXCLUSIVE_LOCK:
                self.__state = _UNLOCKED
            else:
                raise ValueError(f"Not releasable (state {self.__state})")
            #
            self._cond.notify(1)
        #

    def release_shared(self):
        """Release a shared lock"""
        with self._cond:
            if self.__state > 0:
                self.__state -= 1
            else:
                raise ValueError(f"Not releasable (state {self.__state})")
            #
            self._cond.notify(1)
        #


# pylint: disable=too-few-public-methods ; minimal dummy classes


class DummyLockManager:
    """Context manager for DummyLock classes"""

    def __enter__(self) -> bool:
        """Do nothing except returning True"""
        return True

    def __exit__(self, t, v, tb):
        """Do nothing"""


class DummyLock:
    """Dummy lock with shared and exclusive context managers
    as a drop-in replacement for BoltTypeLock in immutable
    wrappers.BaseStructure subclasses that do not require locking
    """

    def __init__(self) -> None:
        """Initialize the value"""
        self.shared = DummyLockManager()
        self.exclusive = DummyLockManager()


# vim: fileencoding=utf-8 ts=4 sts=4 sw=4 autoindent expandtab syntax=python:



# -*- coding: utf-8 -*-
# mypy: disable-error-code="index"

"""
Datastructure wrappers
"""

# import logging

from collections.abc import ItemsView, Iterator, Mapping, MutableMapping
from typing import Iterable, TypeVar, Union, overload, final

from .commons import (
    DOT,
    IndexType,
    SegmentsTuple,
    ValueType,
    assured_iterable,
    assured_index,
)
from .cache import ParsingCache, SINGLETON_CACHES
from .components import Address, get_node, EMPTY_ADDRESS
from .inner_structures import ComponentBasedStructure
from .locks import BoltTypeLock, DummyLock
from .nested_collections import iter_native, iter_nodes


__all__ = ["BaseStructure", "FrozenStructure", "MutableStructure"]

_T = TypeVar("_T")
_NO_DEFAULT = object()

# StructureDataType: TypeAlias = Union[ValueType, "BaseStructure"]
# StructureSourceItem: TypeAlias = tuple[IndexType, ValueType]


class BaseStructure:
    """A data structure, base class for mutable and immutable data structure classes"""

    def __init__(
        self,
        source: Union["BaseStructure", dict[IndexType, ValueType], Iterable] = (),
        /,
        separator: str = DOT,
    ) -> None:
        """Buikd abd store an internal ComponentBasedStructure
        mapping Node intnaces to Address instances
        """
        if isinstance(source, BaseStructure):
            source_iterable: Iterable[tuple[IndexType, ValueType]] = (
                source.stored_items()
            )
        elif isinstance(source, dict):
            source_iterable = source.items()
        elif isinstance(source, Iterable) and not isinstance(source, str):
            source_iterable = assured_iterable(source)
        else:
            raise ValueError(
                "Not a valid source,"
                f" maybe try {self.__class__.__name__}.from_native({source!r})"
            )
        #
        self.__addresses_cache: ParsingCache = SINGLETON_CACHES[separator]
        self.__lock: BoltTypeLock | DummyLock = (
            BoltTypeLock() if self.is_mutable else DummyLock()
        )
        with self.__lock.exclusive:
            self.__internal_cbs = ComponentBasedStructure(
                (self.__addresses_cache[assured_index(index)], get_node(value))
                for index, value in source_iterable
            )
        #

    @classmethod
    def from_native(
        cls, original: ValueType, /, separator: str = DOT
    ) -> "BaseStructure":
        """Create an instance from the native value original"""
        return cls(iter_native(original), separator=separator)

    def to_native(self) -> ValueType:
        """native data"""
        return self.__get_native_data()

    def __get_native_data(self, search_address: Address = EMPTY_ADDRESS) -> ValueType:
        """native data at address Address"""
        with self.__lock.shared:
            try:
                return self.__internal_cbs.get_native_data(search_address)
            except KeyError as error:
                raise KeyError(search_address.segments) from error
            #
        #

    @property
    def is_mutable(self) -> bool:
        """Return True only from on mutable subclasses"""
        raise NotImplementedError

    @property
    def separator(self) -> str:
        """Return the separator"""
        return self.__addresses_cache.separator

    def __getitem__(self, key: IndexType) -> ValueType:
        """Return the substructure determined by full_path"""
        return self.__get_native_data(self.__addresses_cache[key])

    def __repr__(self) -> str:
        """String representation"""
        return (
            f"{self.__class__.__name__}({repr(self.__as_stored_native_dict())},"
            f" separator={self.separator!r})"
        )

    @overload
    def get(self, key: IndexType, /) -> ValueType: ...
    @overload
    def get(self, key: IndexType, /, default: ValueType) -> ValueType: ...
    @overload
    def get(self, key: IndexType, /, default: _T) -> ValueType | _T: ...
    @final
    def get(self, key: IndexType, default=None) -> ValueType:
        """Return the native substructure determined by key"""
        try:
            return self.__get_native_data(self.__addresses_cache[key])
        except KeyError:
            return default
        #

    def __eq__(self, other) -> bool:
        """Equality test"""
        return all(
            (
                self.stored_items() == other.stored_items(),
                self.is_mutable == other.is_mutable,
                self.separator == other.separator,
            )
        )

    def iter_stored(self) -> Iterator[SegmentsTuple]:
        """Iterator over stored indexes"""
        return iter(address.segments for address in self.__internal_cbs)

    def _iter_all_addresses(self) -> Iterator[Address]:
        """Iterator over all addresses: stored and interpolated ones"""
        stored_addresses = list(self.__internal_cbs)
        yield from stored_addresses
        seen_addresses = set(stored_addresses)
        for single_stored_address in stored_addresses:
            current_address = single_stored_address
            while True:
                try:
                    current_address = current_address.parent
                except ValueError:
                    break
                #
                if current_address not in seen_addresses:
                    yield current_address
                #
                seen_addresses.add(current_address)
            #
        #

    def __contains__(self, key, /) -> bool:
        """Return True if key is contained in either stored or interpolated form"""
        address = self.__addresses_cache[key]
        for current_address in self._iter_all_addresses():
            if address == current_address:
                return True
            #
        #
        return False

    def __iter__(self) -> Iterator[SegmentsTuple]:
        """Iterator over (virtually available) indexes"""
        return iter(address.segments for address in self._iter_all_addresses())

    def __len__(self) -> int:
        """Number of stored keys"""
        return len(self.__internal_cbs)

    def __as_stored_native_dict(self) -> dict[SegmentsTuple, ValueType]:
        """Native dict from the cross struct"""
        with self.__lock.shared:
            native_dict = {
                address.segments: node.value
                for address, node in self.__internal_cbs.items()
            }
        #
        return native_dict

    def __as_full_native_dict(self) -> dict[SegmentsTuple, ValueType]:
        """Full native dict from the cross struct"""
        return {
            address.segments: self.__internal_cbs.get_native_data(address)
            for address in self._iter_all_addresses()
        }

    def stored_items(self) -> ItemsView[SegmentsTuple, ValueType]:
        """Iterator over (segments, value) tuples"""
        return self.__as_stored_native_dict().items()

    def items(self) -> ItemsView[SegmentsTuple, ValueType]:
        """Iterator over (segments, value) tuples"""
        return self.__as_full_native_dict().items()

    # TODO: list methods (append, extend, insert)

    def _set_item(self, key, value) -> None:
        """set an item"""
        if not self.is_mutable:
            raise TypeError("Immutable instances do not support item setting")
        #
        with self.__lock.exclusive:
            base_address = self.__addresses_cache[key]
            # Snapshot the data structure first,
            # then try to add all nodes created from the data structure
            with self.__internal_cbs.transaction:
                if value and isinstance(value, (dict, list)):
                    for offset_address, node in iter_nodes(value):
                        self.__internal_cbs[base_address + offset_address] = node
                    #
                #
                else:
                    self.__internal_cbs[base_address] = get_node(value)
                #
            #
        #

    def __or__(self, other: "BaseStructure"):
        """instance | other implementation via an
        intermediate MutableStructure instance
        """
        intermediate_mutable_instance = self.to_mutable_structure()
        intermediate_mutable_instance._update_from(other)
        if self.is_mutable:
            return intermediate_mutable_instance
        #
        return intermediate_mutable_instance.to_frozen_structure()

    def _update_from(self, other: "BaseStructure") -> None:
        """Add the stored items of other"""
        if not self.is_mutable:
            raise TypeError("Immutable instances do not support in-place update")
        #
        with self.__lock.exclusive:
            # Snapshot the data structure first,
            # then try to add all nodes created from the data structure
            with self.__internal_cbs.transaction:
                for segments, value in other.stored_items():
                    self.__internal_cbs[self.__addresses_cache[segments]] = get_node(
                        value
                    )
                #
            #
        #

    def _delete_item(self, key, /) -> None:
        """delete an item"""
        if not self.is_mutable:
            raise TypeError("Immutable instances do not support item deletion")
        #
        with self.__lock.exclusive, self.__internal_cbs.transaction:
            try:
                del self.__internal_cbs[self.__addresses_cache[key]]
            except KeyError as error:
                raise KeyError(key) from error
            #
        #

    @overload
    def _pop(self, key: IndexType, /) -> ValueType: ...
    @overload
    def _pop(self, key: IndexType, /, default: ValueType) -> ValueType: ...
    @overload
    def _pop(self, key: IndexType, /, default: _T) -> ValueType: ...
    @final
    def _pop(self, key: IndexType, default=_NO_DEFAULT) -> ValueType:
        """Return the value determined by key, and delete it"""
        if not self.is_mutable:
            raise TypeError("Immutable instances do not support item popping")
        #
        with self.__lock.exclusive, self.__internal_cbs.transaction:
            cropped = self.__internal_cbs.cropped_substructure(
                self.__addresses_cache[key]
            )
        #
        if not cropped:
            if (
                isinstance(default, (str, int, float, bool, dict, list))
                or default is None
            ):
                return default
            #
            raise KeyError(key)
        #
        return cropped.get_native_data()

    def to_frozen_structure(self) -> "FrozenStructure":
        """Return a FrozenStructure instance"""
        return FrozenStructure(self, separator=self.separator)

    def to_mutable_structure(self) -> "MutableStructure":
        """Return a MutableStructure instance"""
        return MutableStructure(self, separator=self.separator)


class FrozenStructure(BaseStructure, Mapping[IndexType, ValueType]):
    """An immutable data structure"""

    @property
    def is_mutable(self) -> bool:
        """Return True only from on mutable subclasses"""
        return False

    @classmethod
    def from_native(
        cls, original: ValueType, /, separator: str = DOT
    ) -> "FrozenStructure":
        """Create an instance from the native value original"""
        return cls(iter_native(original), separator=separator)

    def copy(self) -> "FrozenStructure":
        """(deep) copy of the instance"""
        return self.to_frozen_structure()

    def __ior__(self, unused_other: BaseStructure):
        """Raies a TyoeError explicitly"""
        raise TypeError("FrozenStructure instances do not support in-place update")


class MutableStructure(BaseStructure, MutableMapping[IndexType, ValueType]):
    """A mutable data structure"""

    @property
    def is_mutable(self) -> bool:
        """Return True on mutable instances"""
        return True

    @classmethod
    def from_native(
        cls, original: ValueType, /, separator: str = DOT
    ) -> "MutableStructure":
        """Create an instance from the native value original"""
        return cls(iter_native(original), separator=separator)

    def copy(self) -> "MutableStructure":
        """(deep) copy of the instance"""
        return self.to_mutable_structure()

    def __setitem__(self, key: IndexType, value: ValueType) -> None:
        """implementation of 'instance[key] = value'"""
        self._set_item(key, value)

    def __delitem__(self, key: IndexType) -> None:
        """implementation of 'del instance[key]'"""
        self._delete_item(key)

    # pylint:disable=arguments-differ ; ok for the type checkers
    @overload
    def pop(self, key: IndexType, /) -> ValueType: ...
    @overload
    def pop(self, key: IndexType, /, default: ValueType) -> ValueType: ...
    @overload
    def pop(self, key: IndexType, /, default: _T) -> ValueType | _T: ...
    @final
    def pop(self, key: IndexType, default=_NO_DEFAULT) -> ValueType:
        """Return the value determined by key, and delete it"""
        return self._pop(key, default=default)

    def __ior__(self, other: "BaseStructure") -> "MutableStructure":
        """Add the stored items of other"""
        self._update_from(other)
        return self


# vim: fileencoding=utf-8 ts=4 sts=4 sw=4 autoindent expandtab syntax=python:



# -*- coding: utf-8 -*-
# mypy: disable-error-code="index"

"""
all things directly related to collections nesting
"""

import logging

from collections.abc import Iterator, Sequence

from .commons import (
    ScalarType,
    ValueType,
    SegmentsTuple,
    assured_int,
    assured_collection,
)
from .components import (
    CompoundKeyError,
    NoMutableCollectionError,
    Address,
    Node,
    ScalarNode,
    ListNode,
    Item,
    EMPTY_ADDRESS,
    get_node,
)


class RecursionDetected(Exception):
    """Raised when a recursion was detected"""

    def __init__(self, message: str):
        """Store the message"""
        self.message = message


class RecursionCircuitBreaker:
    """Circuit breaker for avoiding infinite loops/recursions"""

    def __init__(self, placeholder: ScalarType = None, initial_value: ValueType = None):
        """Store the placeholder"""
        self.__placeholder: ScalarType = placeholder
        self.__registered_ids: list[int] = []
        self.register_id(initial_value)

    def checked_collection_id(
        self, value: ValueType, registry: Sequence[int], verbose: bool = False
    ) -> int:
        """Check if the id of a dict or list is not registered yet.
        Raise an exception if it is.
        """
        if isinstance(value, (dict, list)):
            current_id = id(value)
            if current_id in registry:
                details = f", replacing it by {self.__placeholder!r}" if verbose else ""
                raise RecursionDetected(f"ignoring {value!r} nested in itself{details}")
            #
            return current_id
        #
        raise NoMutableCollectionError

    def register_id(self, value: ValueType, verbose: bool = False) -> None:
        """Register the id of dict or list instances only after checking it"""
        try:
            current_id = self.checked_collection_id(
                value, self.__registered_ids, verbose=verbose
            )
        except NoMutableCollectionError:
            ...
        else:
            self.__registered_ids.append(current_id)
        #

    @property
    def placeholder(self) -> ScalarType:
        """Return the configured placeholder"""
        return self.__placeholder


def get_collection_member(collection: ValueType, key: ScalarType) -> ValueType:
    """Get the member of a dict or list"""
    if isinstance(collection, (dict, list)):
        return collection[key]  # type: ignore[index]
    #
    raise NoMutableCollectionError(
        "Not a collection in the sense of this package,"
        f" cannot retrieve {collection!r}[{key!r}]"
    )


def permissive_partial_traverse(
    start: ValueType,
    address: Address,
    min_remaining_segments: int = 0,
    recursion_placeholder: ScalarType = None,
) -> tuple[ValueType, Address, Address]:
    """Traverse through a data structure starting at the start node,
    until minimum min_remaining_segments of the path are left
    """
    if min_remaining_segments < 0:
        raise ValueError("No negative value allowed here")
    #
    breaker = RecursionCircuitBreaker(
        placeholder=recursion_placeholder, initial_value=start
    )
    pointer = start
    consumed_segments: list[ScalarType] = []
    remaining_segments: list[ScalarType] = list(address.segments)
    while len(remaining_segments) > min_remaining_segments:
        key = remaining_segments.pop(0)
        try:
            pointer = get_collection_member(pointer, key)
        except (IndexError, KeyError, TypeError, NoMutableCollectionError):
            return (
                pointer,
                Address(*consumed_segments),
                Address(key, *remaining_segments),
            )
        #
        consumed_segments.append(key)
        try:
            breaker.register_id(pointer, verbose=True)
        except RecursionDetected as recursion_error:
            logging.warning(str(recursion_error))
            return (
                breaker.placeholder,
                Address(*consumed_segments),
                Address(*remaining_segments),
            )
        #
    #
    return pointer, Address(*consumed_segments), Address(*remaining_segments)


def full_traverse(
    start: ValueType,
    address: Address,
    recursion_placeholder: ScalarType = None,
) -> ValueType:
    """Traverse through a data structure starting at the start node"""
    breaker = RecursionCircuitBreaker(
        placeholder=recursion_placeholder, initial_value=start
    )
    pointer = start
    consumed_segments: list[ScalarType] = []
    remaining_segments: list[ScalarType] = list(address.segments)
    while remaining_segments:
        key = remaining_segments.pop(0)
        consumed_segments.append(key)
        try:
            pointer = get_collection_member(pointer, key)
        except (IndexError, KeyError) as error:
            raise CompoundKeyError(*consumed_segments) from error
        #
        try:
            breaker.register_id(pointer, verbose=True)
        except RecursionDetected as recursion_error:
            logging.warning(str(recursion_error))
            return breaker.placeholder
        #
    #
    return pointer


def traverse_with_default(
    start: ValueType,
    address: Address,
    default: ValueType = None,
) -> ValueType:
    """Traverse through a data structure starting at the start node
    and return the result or the default
    """
    try:
        return full_traverse(start, address)
    except CompoundKeyError:
        return default
    #


def _inner_iter_native(
    breaker: RecursionCircuitBreaker,
    start: ValueType,
    *previous_collection_ids: int,
    previous_segments: SegmentsTuple = (),
) -> Iterator[tuple[SegmentsTuple, ValueType]]:
    """Return an iterator over all SegmentTuples and values"""
    # recursion circuit breaker: in straight line only
    try:
        current_collection_id = breaker.checked_collection_id(
            start, previous_collection_ids, verbose=True
        )
    except RecursionDetected as recursion_error:
        logging.warning(str(recursion_error))
        yield previous_segments, breaker.placeholder
        return
    except NoMutableCollectionError:
        yield previous_segments, start
        return
    #
    subitems: list[tuple[ScalarType, ValueType]] = []
    if isinstance(start, dict):
        subitems = list(start.items())
        if not subitems:
            yield previous_segments, {}
        #
    elif isinstance(start, list):
        subitems = list(enumerate(start))
        yield previous_segments, []
    #
    for key, value in subitems:
        yield from _inner_iter_native(
            breaker,
            value,
            *previous_collection_ids,
            current_collection_id,
            previous_segments=(*previous_segments, key),
        )
    #


def iter_native(
    start: ValueType,
    previous_segments: SegmentsTuple = (),
    recursion_placeholder: ScalarType = None,
) -> Iterator[tuple[SegmentsTuple, ValueType]]:
    """Return an iterator over all native addresses and values
    required to represent a data structure
    """
    breaker = RecursionCircuitBreaker(placeholder=recursion_placeholder)
    yield from _inner_iter_native(
        breaker,
        start,
        previous_segments=previous_segments,
    )


def iter_nodes(
    start: ValueType,
    previous_address: Address = Address(),
    recursion_placeholder: ScalarType = None,
) -> Iterator[tuple[Address, Node]]:
    """Return an iterator over all addressable nodes in a data structure"""
    for segments, value in iter_native(
        start,
        previous_segments=previous_address.segments,
        recursion_placeholder=recursion_placeholder,
    ):
        yield Address(*segments), get_node(value)


def iter_collection_items(
    start: ValueType,
    previous_address: Address = Address(),
) -> Iterator[Item]:
    """Return an iterator over all addressable endpoint items in a data structure"""
    for address, node in iter_nodes(start, previous_address=previous_address):
        yield Item(address, node)
    #


def iter_addresses(
    start: ValueType,
    previous_address: Address = Address(),
) -> Iterator[Address]:
    """Return an iterator over all endpoint paths in the data structure"""
    for address, _ in iter_nodes(start, previous_address=previous_address):
        yield address
    #


def construct(crosswise: dict[SegmentsTuple, ValueType]) -> ValueType:
    """construct a nested collection from a mapping"""
    if not crosswise:
        raise ValueError(
            "Constructing a nested collection from an empty dict is not supported"
        )
    #
    base_value = crosswise.pop((), {})
    if not crosswise:
        return base_value
    #
    child_nodes: dict[ScalarType, list[SegmentsTuple]] = {}
    for subaddress in crosswise:
        current_key = subaddress[0]
        child_nodes.setdefault(current_key, []).append(subaddress)
    #
    if isinstance(base_value, list):
        indexes: list[int] = [-1]
        indexes.extend(map(assured_int, child_nodes))
        for _ in range(max(indexes) + 1):
            base_value.append(None)
        #
    elif not isinstance(base_value, dict):
        logging.debug(
            "Traversing through a leaf, overriding value %r",
            base_value,
        )
        base_value = {}
    #
    base_value = assured_collection(base_value)
    for key, subkeys in child_nodes.items():
        subcrosswise: dict[SegmentsTuple, ValueType] = {
            segments[1:]: crosswise.pop(segments) for segments in subkeys
        }
        base_value[key] = construct(subcrosswise)  # ty: ignore[invalid-assignment]
    #
    if crosswise:
        # should never occur, so not really testable
        raise ValueError(
            f"Unexpected: not all addresses consumed, remaining: {crosswise!r}"
        )
    #
    return base_value


def get_child_addresses(
    components_mapping: dict[Address, Node], key: Address
) -> list[Address]:
    """Return a list of child addresses of components_mapping[key]"""
    key_length = len(key)
    return [
        child_address
        for child_address in components_mapping
        if child_address.startswith(key) and len(child_address) > key_length
    ]


def cleanup_after_removal(
    components_mapping: dict[Address, Node], removed_key: Address
) -> None:
    """Do cleanup in a components mapping after removed_key has been deleted
    from it, modifying components_mapping in place
    """
    # cleanup of collection nodes whose children have all been removed
    # (by removing removed_key)
    current_ancestor_address = removed_key
    # pylint: disable=too-many-nested-blocks ; refactoring candidate
    while True:
        try:
            current_ancestor_address = current_ancestor_address.parent
        except ValueError:
            break
        #
        if current_ancestor_address not in components_mapping:
            continue
        #
        children = get_child_addresses(components_mapping, current_ancestor_address)
        if children:
            # logging.warning("Fixing list %r", children)
            # Rearrange list node children if required,
            # e.g. if a list contained 6 elements,
            #      and index 2 got deleted, then:
            #           [0] [1] [2] [3] [4] [5]                 (original)
            #           [0] [1] [X] [3] {4] [5]                 (deletion)
            #           [0→0] [1→1] [2→void] [3→2] [4→3] [5→4]  (re-map)
            #           [0] [1] [2] [3] [4]                     (result)
            current_address_length = len(current_ancestor_address)
            if isinstance(components_mapping[current_ancestor_address], ListNode):
                indexes = sorted(
                    set(
                        assured_int(child_address[current_address_length])
                        for child_address in children
                    )
                )
                key_subindex = assured_int(removed_key[current_address_length])
                if key_subindex not in indexes and key_subindex < max(indexes):
                    for child_address in children:
                        child_subindex = assured_int(
                            child_address[current_address_length]
                        )
                        if child_subindex > key_subindex:
                            new_segments = list(child_address.segments)
                            new_segments[current_address_length] = child_subindex - 1
                            # no .pop() here for type safety
                            saved_node = components_mapping[child_address]
                            del components_mapping[child_address]
                            components_mapping[Address(*new_segments)] = saved_node
                        #
                    #
                #
            #
            # Exit the loop if any children were found, no further cleanup necessary
            break
        #
        del components_mapping[current_ancestor_address]
    #


def check_ancestors_before_adding_address(
    components_mapping: dict[Address, Node], add_candidate: Address
) -> None:
    """Check all ancestors if the added key fits the list"""
    current_key = add_candidate
    while True:
        try:
            current_key = current_key.parent
        except ValueError:
            break
        #
        if current_key not in components_mapping:
            continue
        #
        current_value = components_mapping[current_key]
        if isinstance(current_value, ScalarNode):
            # Remove any ScalarNode at an ancestor position
            del components_mapping[current_key]
        elif isinstance(current_value, ListNode):
            current_key_length = len(current_key)
            significant_index = assured_int(add_candidate[current_key_length])
            if significant_index < 0:
                raise IndexError("Cannot use a negative index here")
            #
            children = get_child_addresses(components_mapping, current_key)
            existing_indexes = sorted(
                set(
                    assured_int(child_address[current_key_length])
                    for child_address in children
                )
            )
            if existing_indexes:
                pos_after_list = significant_index - max(existing_indexes)
                if pos_after_list > 1:
                    raise ValueError(
                        "Implicitly extending a list by more than one item"
                        " is not implemented"
                    )
                #
                if pos_after_list == 1:
                    logging.info("Implicitly extending list by one item")
                #
            #
        #
        break
    #


def matching_items(
    components_mapping: dict[Address, Node],
    /,
    search_address: Address = EMPTY_ADDRESS,
) -> list[tuple[Address, Node]]:
    """Get a list of matching item tuples"""
    return [
        (stored_address, stored_node)
        for (stored_address, stored_node) in components_mapping.items()
        if stored_address.startswith(search_address)
    ]


# vim: fileencoding=utf-8 ts=4 sts=4 sw=4 autoindent expandtab syntax=python:



# -*- coding: utf-8 -*-

"""
ParsingCache
"""

from collections import Counter
from collections.abc import Mapping
from json import JSONDecodeError, dumps, loads
from re import Match, compile as re_compile
from threading import Lock

from .commons import DOT, DOUBLE_QUOTE, EMPTY, IndexType, ScalarType
from .components import Address


NEXT_QUOTED = re_compile('^"([^"]+)"')
NEXT_IN_SUBSCRIPT_QUOTED = re_compile(r'^\["([^"]+)"\]')
NEXT_IN_SUBSCRIPT_SIMPLE = re_compile(r"^\[([^\]]+)\]")


class OneCharSeparated:
    """Base class having a separator property given at init"""

    def __init__(self, separator: str = DOT) -> None:
        """Initialize with a single-character separator"""
        if len(separator) != 1:
            raise TypeError("The separator must be a single character")
        #
        self.__separator = separator

    @property
    def separator(self) -> str:
        """Return the separator string"""
        return self.__separator

    def __repr__(self) -> str:
        """String representation"""
        return f"{self.__class__.__name__}(separator={self.separator!r})"


class SegmentsParser(OneCharSeparated):
    """Can build a tuple of segments from a string"""

    def __init__(self, separator: str = DOT) -> None:
        """Initialize with a single-character separator"""
        super().__init__(separator=separator)
        self._active = False
        self._expect_segment_end = False
        self._collected_segments: list[ScalarType] = []
        self._current_segment_sources: list[str] = []

    def _add_segment(self, segment_source: str) -> None:
        """Add a new segment from segment_source"""
        try:
            segment = loads(segment_source)
        except JSONDecodeError:
            segment = segment_source
        #
        self._collected_segments.append(segment)
        self._current_segment_sources.clear()

    def _store_and_reset_segment(self) -> None:
        """reset the internal state between segments"""
        if self._current_segment_sources:
            self._add_segment(EMPTY.join(self._current_segment_sources))
        #
        self._expect_segment_end = False

    def _add_match_and_get_end_pos(self, match: Match, quote: bool = False) -> int:
        """Add the matched portion that is quoted or in a subscript ([...])
        and return the match end position
        """
        self._expect_segment_end = True
        segment_source = f'"{match.group(1)}"' if quote else match.group(1)
        self._add_segment(segment_source)
        return match.end()

    def _check_for_fast_forward(self, path_source: str, pos: int) -> int:
        """Check if we can fast-forward"""
        if self._current_segment_sources:
            return 0
        #
        remainder = path_source[pos:]
        character = remainder[0]
        if character == DOUBLE_QUOTE:
            match = NEXT_QUOTED.match(remainder)
            if match:
                return self._add_match_and_get_end_pos(match, quote=True)
            #
        elif character == "[":
            quote = True
            match = NEXT_IN_SUBSCRIPT_QUOTED.match(remainder)
            if not match:
                quote = False
                match = NEXT_IN_SUBSCRIPT_SIMPLE.match(remainder)
            #
            if match:
                return self._add_match_and_get_end_pos(match, quote=quote)
            #
        #
        return 0

    def split_into_segments(self, path_source: str) -> Address:
        """Split a string into an Address instance suitable
        for addressing an item in a netsted collection
        """
        if self._active:
            raise ValueError(
                f"{self.__class__.__name__} instances are not thread-safe,"
                " concurrent execution on the same instance is not supported."
            )
        #
        self._active = True
        self._store_and_reset_segment()
        self._collected_segments.clear()
        pos = 0
        path_source_length = len(path_source)
        while pos < path_source_length:
            character = path_source[pos]
            if character == self.separator:
                self._store_and_reset_segment()
                pos += 1
                continue
            #
            if self._expect_segment_end:
                raise ValueError(
                    f"Expected segment end but read character {character!r}."
                    f" Collected segments so far: {self._collected_segments!r}"
                )
            #
            fast_forward = self._check_for_fast_forward(path_source, pos)
            if fast_forward:
                pos += fast_forward
            else:
                self._current_segment_sources.append(character)
                pos += 1
            #
        #
        self._store_and_reset_segment()
        found_segments = tuple(self._collected_segments)
        self._collected_segments.clear()
        self._active = False
        return Address(*found_segments)


class ParsingCache(OneCharSeparated):
    """Can build a tuple of segments from a string"""

    def __init__(self, separator: str = DOT) -> None:
        """Initialization argument:
        * separator - a single character used to separate path segments
        """
        super().__init__(separator=separator)
        self.__cache: dict[str, Address] = {}
        self.stats: Counter[str] = Counter()
        self.__lock = Lock()

    def __getitem__(self, pathspec: IndexType) -> Address:
        """Return an item from the cache if the pathspec is a string"""
        if isinstance(pathspec, str):
            return self.get_cached(pathspec)
        #
        self.stats.update(bypass=1)
        return Address(*pathspec)

    def get_cached(self, path_source: str) -> Address:
        """Return an item from the cache"""
        if not path_source:
            self.stats.update(bypass=1)
            return Address()
        #
        with self.__lock:
            stat_entries = []
            try:
                cached_address = self.__cache[path_source]
            except KeyError:
                parser = SegmentsParser(separator=self.separator)
                cached_address = self.__cache.setdefault(
                    path_source, parser.split_into_segments(path_source)
                )
                stat_entries.append("miss")
            else:
                stat_entries.append("hit")
            #
            self.stats.update(stat_entries)
            return cached_address
        #

    def canonical(self, source_address: Address) -> str:
        """Return the canonical source for the given segments tuple"""
        output: list[str] = []
        for item in source_address.segments:
            item_dump = dumps(item)
            if isinstance(item, str) and not any(
                char in item for char in (self.separator, DOUBLE_QUOTE, "[")
            ):
                item_dump = str(item)
            elif isinstance(item, (int, float)):
                # Number representations might contain the separator character
                if self.separator in item_dump:
                    item_dump = f"[{item_dump}]"
                #
            #
            output.append(item_dump)
        #
        reconstructed_key = self.separator.join(output)
        with self.__lock:
            cached_address = self.__cache.setdefault(reconstructed_key, source_address)
        #
        if cached_address != source_address:
            # should never occur, but added as consistency check
            raise ValueError("Unexpected: canonical representation mismatch")
        #
        return reconstructed_key


class CacheOfCaches(Mapping):
    """A cache of ParsingCache instances"""

    def __init__(self) -> None:
        """Initialize the internal cache"""
        self.__dot_cache = ParsingCache(separator=DOT)
        self.__other_caches: dict[str, ParsingCache] = {}

    def __getitem__(self, name: str) -> ParsingCache:
        """Return the matching ParsingCache, or create a new one"""
        if name == DOT:
            return self.__dot_cache
        #
        try:
            return self.__other_caches[name]
        except KeyError:
            return self.__other_caches.setdefault(name, ParsingCache(separator=name))
        #

    def __iter__(self):
        """Iterate over the keys"""
        yield DOT
        yield from self.__other_caches
        # keys = [DOT] + list(self.__other_caches)
        # return iter(keys)

    def __len__(self) -> int:
        """Number of stored caches"""
        return len(self.__other_caches) + 1


SINGLETON_CACHES = CacheOfCaches()


# vim: fileencoding=utf-8 ts=4 sts=4 sw=4 autoindent expandtab syntax=python:



# -*- coding: utf-8 -*-

"""
Call as script
"""

import logging
import sys

from argparse import REMAINDER, ArgumentParser
from enum import Enum
from importlib.metadata import PackageNotFoundError, version as metadata_version
from json import dumps as json_dumps, load as json_load
from pathlib import Path
from sys import stdin

from . import FrozenStructure
from .commons import DOT

if sys.version_info >= (3, 11):
    from tomllib import loads as tomli_loads
else:
    from tomli import loads as tomli_loads  # ty: ignore[unresolved-import]
#


_PACKAGE_NAME = "ncw"
_DEFAULT_SEPARATOR = DOT
_INFO_TEXT = """This package does not have a useful commandline interface;
programs like jq and yq offer much more sophisticated functionality
for shell environments, and we do not intend to re-invent the wheel.

That being said, you can use the '-c' option to construct a simple
nested collection using one or more 'deep.nested.locator=value' expressions
and prion it to stdout as JSON,
or the '-x' option to extract values from JSON provided through standard input
using one or more 'deep.nested.locator' expressions, and print the extracted
values to stdout.

Package documentation is available at https://blackstream-x.gitlab.io/ncw/"""


class Action(Enum):
    """Possible actions"""

    NOTHING = 0
    EXTRACT = 1
    CONSTRUCT = 2


def get_file_contents(file_name: str, up_dirs: int = 0) -> str:
    """get contents of a file in an ancestor directory
    of the current file (up_dirs directory levels up)
    """
    file_dir_path = Path(__file__).resolve()
    for _ in range(up_dirs):
        file_dir_path = file_dir_path.parent
    #
    file_path = file_dir_path / file_name
    return file_path.read_text(encoding="utf-8")


def get_metadata_version(
    metadata_file_name: str = "pyproject.toml", up_dirs: int = 3
) -> str:
    """get version information from metadata_file up_dirs directories above"""
    try:
        metadata_file_contents = get_file_contents(metadata_file_name, up_dirs=up_dirs)
    except IOError as error:
        return str(error)
    #
    metadata = FrozenStructure.from_native(tomli_loads(metadata_file_contents))
    try:
        # logging.warning(str(metadata.items()))
        version = metadata["project.version"]
    except KeyError:
        return f"Error: no version information in metadata from {metadata_file_name}"
    #
    return f"{version} (read directly from {metadata_file_name})"


def construct(expressions: list[str], separator: str = _DEFAULT_SEPARATOR) -> None:
    """Construct a structure and print its JSON representation"""
    source_items: list[tuple[str, str]] = []
    for index, current_expression in enumerate(expressions):
        try:
            collection_index, value = current_expression.split("=", 1)
        except ValueError as error:
            raise ValueError(
                f"Expression #{index + 1} ({current_expression!r})"
                " must contain at least one 'equals' sign ('=')"
            ) from error
        #
        source_items.append((collection_index, value))
    #
    fs = FrozenStructure(source_items, separator=separator)
    print(json_dumps(fs.to_native(), indent=2))


def extract(expressions: list[str], separator: str = _DEFAULT_SEPARATOR) -> None:
    """Extract items from the nested collection receibed as JSON through stdin"""
    fs = FrozenStructure.from_native(
        json_load(stdin),
        separator=separator,
    )
    for collection_index in expressions:
        value = fs[collection_index]
        if isinstance(value, str):
            print(value)
        else:
            print(json_dumps(value, indent=2))
        #
    #


def dispatch(
    action: Action, expressions: list[str], separator: str = _DEFAULT_SEPARATOR
) -> None:
    """Dispatch to the specialized functions"""
    if expressions:
        if action == Action.CONSTRUCT:
            return construct(expressions, separator=separator)
        #
        if action == Action.EXTRACT:
            return extract(expressions, separator=separator)
        #
    #
    return print(_INFO_TEXT)


def main() -> int:
    """Script fuction autogenerated by uv, might delete later"""
    try:
        version = metadata_version(_PACKAGE_NAME)
    except PackageNotFoundError:
        version = get_metadata_version()
    #
    argument_parser = ArgumentParser(
        prog="ncw", description="Nested collections wrapper"
    )
    argument_parser.set_defaults(action=Action.NOTHING)
    argument_parser.add_argument("--version", action="version", version=version)
    argument_parser.add_argument(
        "-s", "--separator", default=".", help="separator character"
    )
    mutex_group = argument_parser.add_mutually_exclusive_group()
    mutex_group.add_argument(
        "-c",
        "--construct",
        action="store_const",
        const=Action.CONSTRUCT,
        dest="action",
        help="construct a nested data structure",
    )
    mutex_group.add_argument(
        "-x",
        "--extract",
        action="store_const",
        const=Action.EXTRACT,
        dest="action",
        help="extract an item from the data structure read from stdin",
    )
    argument_parser.add_argument("expressions", nargs=REMAINDER, help="index=value")
    arguments = argument_parser.parse_args()
    try:
        dispatch(arguments.action, arguments.expressions, separator=arguments.separator)
    except (IndexError, KeyError, TypeError, ValueError) as error:
        logging.error(repr(error))
        return 1
    #
    return 0


if __name__ == "__main__":
    sys.exit(main())  # not testable

# vim: fileencoding=utf-8 ts=4 sts=4 sw=4 autoindent expandtab syntax=python:



# -*- coding: utf-8 -*-

"""
components of nested collections: Addresses, Nodes, and Items
"""

from collections.abc import Sequence
from dataclasses import dataclass
from typing import Any, final, overload

from .commons import ScalarType, COMMA_BLANK, assured_int, assured_scalar


def get_segments_representation(*segments: ScalarType) -> str:
    """Return a representation of the segments"""
    return COMMA_BLANK.join([repr(item) for item in segments])


class NoMutableCollectionError(Exception):
    """Raised when trying to traverse throush a leaf of a nested collection"""


class CompoundKeyError(Exception):
    """Raised with the consumed keys if traversal failed"""

    def __init__(self, *segments) -> None:
        """Store the segments"""
        self.segments = segments

    def __repr__(self) -> str:
        """Representation"""
        return (
            f"{self.__class__.__name__}({get_segments_representation(*self.segments)})"
        )

    def __str__(self) -> str:
        """String value"""
        return repr(self.segments)


class Address(Sequence[ScalarType]):
    """Address in a nested collection, basically a tuple of scalars"""

    def __init__(self, *segments: ScalarType) -> None:
        """Store the segments"""
        self.segments: tuple[ScalarType, ...] = segments

    @overload
    def __getitem__(self, index: int) -> ScalarType: ...
    @overload
    def __getitem__(self, index: slice) -> "Address": ...
    @final
    def __getitem__(self, index: slice | int) -> Any:
        """Direct segments access"""
        if isinstance(index, slice):
            new_segments: tuple[ScalarType, ...] = self.segments[index]
            return Address(*new_segments)
        #
        return assured_scalar(self.segments[assured_int(index)])

    def __repr__(self) -> str:
        """Representation"""
        return (
            f"{self.__class__.__name__}({get_segments_representation(*self.segments)})"
        )

    def __str__(self) -> str:
        """String value"""
        return repr(self.segments)

    def __hash__(self) -> int:
        """hash value"""
        return hash(repr(self))

    def __eq__(self, other) -> bool:
        """rich comparision: equals"""
        return repr(self) == repr(other)

    def __bool__(self) -> bool:
        """bool value"""
        return bool(self.segments)

    def __len__(self) -> int:
        """length"""
        return len(self.segments)

    def __add__(self, other) -> "Address":
        """append other"""
        added_segments = self.segments + other.segments
        return Address(*added_segments)

    @property
    def parent(self) -> "Address":
        """return the parent address"""
        if not self:
            raise ValueError("An empty address has no parent")
        #
        return Address(*self.segments[:-1])

    def startswith(self, other: "Address") -> bool:
        """Return True if the segments begin with other’s segments"""
        if not other:
            return True
        #
        if len(other) > len(self):
            return False
        #
        for index, item in enumerate(other.segments):
            if self.segments[index] != item:
                return False
            #
        #
        return True

    def relative_to(self, other: "Address") -> "Address":
        """return the addtress of self relative to other"""
        if self.startswith(other):
            return Address(*self.segments[len(other) :])
        #
        raise ValueError(f"{self} does not start with {other}")


EMPTY_ADDRESS = Address()


# pylint: disable = too-few-public-methods


class Node:
    """Node in a nested collection"""

    @property
    def value(self):
        """return the value"""
        raise NotImplementedError

    def __eq__(self, other) -> bool:
        """rich comparision: equals"""
        return self.value == other.value


class ScalarNode(Node):
    """Scalar value in a nested collection"""

    def __init__(self, value) -> None:
        """set value"""
        self.__value: ScalarType = assured_scalar(value)

    @property
    def value(self) -> ScalarType:
        """return the value"""
        return self.__value

    def __repr__(self) -> str:
        """Representation"""
        return f"{self.__class__.__name__}({self.value!r})"


class CollectionNode(Node):
    """Collection"""

    @property
    def value(self):
        """return the value"""
        raise NotImplementedError

    def __repr__(self) -> str:
        """Representation"""
        return f"{self.__class__.__name__}()"


class ListNode(CollectionNode):
    """list in a nested collection"""

    @property
    def value(self):
        """return an empty list"""
        return []


class MappingNode(Node):
    """mapping (ie. dict) in a nested collection"""

    @property
    def value(self):
        """return an empty dict"""
        return {}


LIST_NODE = ListNode()
MAPPING_NODE = MappingNode()


@dataclass(frozen=True)
class Item:
    """Item in a collection"""

    address: Address
    node: Node


def get_node(value: Any) -> Node:
    """Return a node from the original value"""
    if isinstance(value, list):
        return LIST_NODE
    #
    if isinstance(value, dict):
        return MAPPING_NODE
    #
    return ScalarNode(value)


def all_positive_indexes_at_position(addresses: list[Address], position: int) -> bool:
    """Return True if all addresses in the list have a positive index number
    at the specified position
    """
    for single_address in addresses:
        try:
            if assured_int(single_address[position]) < 0:
                return False
            #
        except (IndexError, TypeError):
            # IndexError: Address to short
            # TypeError: single_address[position] is not an int
            return False
        #
    #
    return True


def assured_node(
    original: Any, error_message: str = "Expected a Node instance"
) -> Node:
    """assure the original is a Node instance"""
    if isinstance(original, Node):
        return original
    #
    raise TypeError(error_message)


# vim: fileencoding=utf-8 ts=4 sts=4 sw=4 autoindent expandtab syntax=python:



# -*- coding: utf-8 -*-
# mypy: disable-error-code="index"

"""
Inner datasructure for the wrappers
"""

import logging

from collections.abc import ItemsView, Iterator, MutableMapping
from typing import Any, Iterable

from .commons import ValueType, assured_int
from .components import (
    Address,
    Node,
    ScalarNode,
    ListNode,
    all_positive_indexes_at_position,
    EMPTY_ADDRESS,
    MAPPING_NODE,
)
from .nested_collections import (
    construct,
    cleanup_after_removal,
    get_child_addresses,
    check_ancestors_before_adding_address,
    matching_items,
)

__all__ = ["ComponentBasedStructure", "Transaction"]

_NOT_PROVIDED = object()


class ComponentBasedStructure(MutableMapping[Address, Node]):
    """A non-threadsafe, mutable datastructure with cross-addressing"""

    def __init__(
        self,
        source: Iterable | "ComponentBasedStructure" | dict[Address, Node] = (),
    ) -> None:
        """build the internal structure"""
        if isinstance(source, (ComponentBasedStructure, dict)):
            data_source: Iterable = list(source.items())
        else:
            data_source = source
        #
        self.__data: dict[Address, Node] = {}
        self.replace_data(data_source)
        self.transaction = Transaction(self)

    def replace_data(self, source: Iterable) -> None:
        """Replace internal data by source"""
        self.__data.clear()
        self.__data.update(source)

    def __getitem__(self, address: Address) -> Node:
        """Return the substructure determined by full_path"""
        return self.__data[address]

    def __delitem__(self, address: Address) -> None:
        """Delete an item"""
        self.delete(address)

    def __setitem__(self, address: Address, value: Node) -> None:
        """set an item"""
        check_ancestors_before_adding_address(self.__data, address)
        #
        # Check descendants
        children = get_child_addresses(self.__data, address)
        if isinstance(value, ScalarNode):
            # ScalarNode instances may not have any children
            for surplus_child_address in children:
                del self.__data[surplus_child_address]
            #
        elif isinstance(value, ListNode):
            # ensure all children have a positive numeric index
            if not all_positive_indexes_at_position(children, len(address)):
                raise ValueError("Existing descendants do not fit a list")
            #
        #
        self.__data[address] = value

    def get_native_data(self, /, search_address: Address = EMPTY_ADDRESS) -> ValueType:
        """Return native data constructed from the stored address, node ..."""
        filtered = matching_items(self.__data, search_address=search_address)
        if not filtered:
            if search_address == EMPTY_ADDRESS:
                return {}
            #
            raise KeyError(search_address)
        #
        return construct(
            {
                found_address.relative_to(search_address).segments: node.value
                for found_address, node in filtered
            }
        )

    def items(self) -> ItemsView[Address, Node]:
        """internally stored items"""
        return self.__data.items()

    def __contains__(self, address, /) -> bool:
        """Number of items"""
        return address in self.__data

    def __iter__(self) -> Iterator[Address]:
        """Iterator over addresses"""
        return iter(self.__data)

    def __len__(self) -> int:
        """Number of items"""
        return len(self.__data)

    def __eq__(self, other) -> bool:
        """Rich comparison: equal items"""
        return self.items() == other.items()

    def get(self, key: Address, default=None) -> Any:
        """Return the Node instance determined by key, or default"""
        try:
            return self.__data[key]
        except KeyError:
            return default
        #

    def _popped_items(
        self,
        address: Address,
        /,
    ) -> Iterator[tuple[Address, Node]]:
        """Return items deleted by the pop operation"""
        filtered = matching_items(self.__data, search_address=address)
        for found_address, node in filtered:
            del self.__data[found_address]
            yield found_address, node
        #

    def cropped_substructure(self, address: Address) -> "ComponentBasedStructure":
        """Return the cropped substructure with addresses relative to address"""
        return ComponentBasedStructure(
            [
                (cropped_address.relative_to(address), node)
                for (cropped_address, node) in self._popped_items(address)
            ]
        )

    def graft(
        self, address: Address, scion: "ComponentBasedStructure", cutoff: bool = False
    ) -> None:
        """modify the instance in place by adding another instance –
        called "scion" here – at address, similar to grafting in horticulture.
        If cutoff is set True, remove all addresses starting with address
        before adding scion.
        This method should preferably called within a transaction
        to ensure a consistent state.
        """
        if not scion:
            return
        #
        if EMPTY_ADDRESS not in scion and address in self.__data:
            if isinstance(self.__data[address], ListNode):
                for sub_address in scion:
                    if sub_address:
                        candidate = sub_address[0]
                        assured_int(
                            candidate,
                            error_message=f"{candidate!r} is not a valid list index",
                        )
                    #
                #
            else:
                del self.__data[address]
            #
        #
        if cutoff:
            list(self._popped_items(address))
        #
        for sub_address, new_node in scion.items():
            self.__data[address + sub_address] = new_node
        #

    # pylint: disable=arguments-renamed ; renamed for consitency
    def pop(self, address: Address, default: Any = _NOT_PROVIDED) -> Any:
        """Return the Node instance determined by address, and delete it
        and its descendants, but do not dot remove any ancestors left childless
        """
        found_items = dict(self._popped_items(address))
        if not found_items:
            if isinstance(default, Node):
                return default
            #
            raise KeyError(address)
        #
        real_default: Node = MAPPING_NODE
        if isinstance(default, Node):
            real_default = default
        #
        found_value: Node = found_items.pop(address, real_default)
        for descendant_address, descendant_value in found_items.items():
            logging.debug("Also removed %r → %r", descendant_address, descendant_value)
        #
        return found_value

    def delete(self, address: Address, /) -> None:
        """Delete an item without returning it"""
        found_items = dict(self._popped_items(address))
        if not found_items:
            raise KeyError(address)
        #
        try:
            del found_items[address]
        except KeyError as error:
            logging.debug("implicit match for %r", str(error))
        #
        cleanup_after_removal(self.__data, address)


class Transaction:
    """Transaction around an update of a ComponentBasedStructure instance"""

    def __init__(self, cbs_instance: ComponentBasedStructure) -> None:
        """Store the ComponentBasedStructure instance"""
        self.__cbs_instance = cbs_instance
        self.__snapshot: tuple[tuple[Address, Node], ...] = ()

    def _draw_snapshot(self) -> None:
        """Store a snapshot of the instance"""
        self.__snapshot = tuple(self.__cbs_instance.items())

    def _restore_snapshot(self) -> None:
        """Resore the instance to the snapshot"""
        self.__cbs_instance.replace_data(self.__snapshot)

    def __enter__(self):
        """Cpontect manager enter: draw a snapshot"""
        self._draw_snapshot()
        return self

    def __exit__(self, exc_type, unused_exc_value, unused_traceback):
        """Context manager exit: Restore the snapshot if anything went wrong"""
        if exc_type is not None:
            self._restore_snapshot()
        #


# vim: fileencoding=utf-8 ts=4 sts=4 sw=4 autoindent expandtab syntax=python:



# -*- coding: utf-8 -*-

"""
simple nested collections wrapper
"""

# pylint: disable=useless-import-alias ; re-import for public interface as
# suggested in <https://docs.astral.sh/ruff/rules/unused-import/#why-is-this-bad>

from .wrappers import (
    BaseStructure as BaseStructure,
    FrozenStructure as FrozenStructure,
    MutableStructure as MutableStructure,
)

from .commons import (
    assured_collection as assured_collection,
    assured_dict as assured_dict,
    assured_float as assured_float,
    assured_int as assured_int,
    assured_iterable as assured_iterable,
    assured_list as assured_list,
    assured_number as assured_number,
    assured_scalar as assured_scalar,
    assured_str as assured_str,
    to_text as to_text,
)


# vim: fileencoding=utf-8 ts=4 sts=4 sw=4 autoindent expandtab syntax=python:
